using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;


namespace org.wayoda.csiow {
    /// <summary>
    /// Contains the definitions and method-calls into unmanaged code
    /// nedded for enumerating the IOWarriors and the DLL-Imports for
    /// read/write operations.
    /// </summary>
    internal sealed class HidDevice {

        // ******************************************************************************
        // Types and constants for file IO
        // ******************************************************************************
        public const uint GENERIC_READ=0x80000000;
        public const uint GENERIC_WRITE=0x40000000;
        public const uint FILE_SHARE_READ=0x00000001;
        public const uint FILE_SHARE_WRITE=0x00000002;
        public const uint FILE_FLAG_OVERLAPPED=0x40000000;
        public const int INVALID_HANDLE_VALUE=-1;
        public const short OPEN_EXISTING=3;
        public const uint WAIT_FAILED=0xFFFFFFFF;
        public const uint WAIT_TIMEOUT=0x102;
        public const uint WAIT_OBJECT_0=0;
        public const int ERROR_HANDLE_EOF=38;

        [StructLayout(LayoutKind.Sequential)]
        public struct SECURITY_ATTRIBUTES {
            public int nLength;
            public int lpSecurityDescriptor;
            public int bInheritHandle;
        }

        // ******************************************************************************
        // Types and constants for enumerating the hid devices
        // ******************************************************************************
        public const int DBT_DEVICEARRIVAL = 0x8000;
        public const int DBT_DEVICEREMOVECOMPLETE = 0x8004;
        public const int DBT_DEVTYP_DEVICEINTERFACE = 5;
        public const int DEVICE_NOTIFY_WINDOW_HANDLE = 0;
        public const int WM_DEVICECHANGE = 0x219;

        public const short DIGCF_PRESENT = 0x00000002;
        public const short DIGCF_DEVICEINTERFACE = 0x00000010;

        [StructLayout(LayoutKind.Sequential)]
        public class DEV_BROADCAST_DEVICEINTERFACE {
            public int dbcc_size;
            public int dbcc_devicetype;
            public int dbcc_reserved;
            public Guid dbcc_classguid;
            public short dbcc_name;
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct SP_DEVICE_INTERFACE_DATA {
            public int cbSize;
            public System.Guid InterfaceClassGuid;
            public int Flags;
            public int Reserved;
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct SP_DEVICE_INTERFACE_DETAIL_DATA {
            public int cbSize;
            public string DevicePath;
        }

        /// <summary>
        /// Needed for testing if an HID is an IOWarrior
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct HIDD_ATTRIBUTES {
            public int Size;
            public short VendorID;
            public short ProductID;
            public short VersionNumber;
        }

        /// <summary>
        /// Needed to check the interface that is represented by a handle
        /// for an IOWarrior
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        public struct HIDP_CAPS {
            public short Usage;
            public short UsagePage;
            public short InputReportByteLength;
            public short OutputReportByteLength;
            public short FeatureReportByteLength;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst=17)]
            public short[] Reserved;
            public short NumberLinkCollectionNodes;
            public short NumberInputButtonCaps;
            public short NumberInputValueCaps;
            public short NumberInputDataIndices;
            public short NumberOutputButtonCaps;
            public short NumberOutputValueCaps;
            public short NumberOutputDataIndices;
            public short NumberFeatureButtonCaps;
            public short NumberFeatureValueCaps;
            public short NumberFeatureDataIndices;
        }

        
        //***************************************************************************
        // functions to be imported for file io on the interfaces of the iowarrior
        //***************************************************************************
        
        [DllImport("kernel32.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern int CreateFile(string lpFileName, uint dwDesiredAccess,
            uint dwShareMode, ref SECURITY_ATTRIBUTES lpSecurityAttributes,
            int dwCreationDisposition, uint dwFlagsAndAttributes, int hTemplateFile);

        [DllImport("kernel32.dll", SetLastError=true)]
        static public extern int ReadFile(int hFile, byte[] lpBuffer,
            int nNumberOfBytesToRead, ref int lpNumberOfBytesRead,
            ref System.Threading.NativeOverlapped lpOverlapped);

        [DllImport("kernel32.dll", SetLastError=true)]
        static public extern int WriteFile(int hFile, byte[] lpBuffer,
            int nNumberOfBytesToWrite, ref int lpNumberOfBytesWritten, 
            ref System.Threading.NativeOverlapped lpOverlapped);
        
        [DllImport("kernel32.dll")]
        static public extern bool GetOverlappedResult(int hFile,
           ref System.Threading.NativeOverlapped lpOverlapped,
           ref int lpNumberOfBytesTransferred, bool bWait);
        
        [DllImport("kernel32.dll", SetLastError=true)]
        static public extern int CancelIo(int hFile);

        [DllImport("kernel32.dll", SetLastError=true)]
        static public extern int CloseHandle(int hObject);

        [DllImport("kernel32.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern int CreateEvent(ref SECURITY_ATTRIBUTES SecurityAttributes,
            int bManualReset, int bInitialState, string lpName);

        [DllImport("kernel32.dll", SetLastError=true)]
        static public extern bool SetEvent(IntPtr hEvent);

        [DllImport("kernel32.dll", SetLastError=true)]
        static public extern uint WaitForMultipleObjects(uint nCount, IntPtr[] lpHandles,
           bool bWaitAll, uint dwMilliseconds);

        //***************************************************************************
        // functions to be imported for device detection
        //***************************************************************************
        /// <summary>
        /// Register a callback for device changes
        /// </summary>
        /// <param name="hRecipient"></param>
        /// <param name="NotificationFilter"></param>
        /// <param name="Flags"></param>
        /// <returns></returns>
        [DllImport("user32.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern IntPtr RegisterDeviceNotification(IntPtr hRecipient,
            IntPtr NotificationFilter, int Flags);

        /// <summary>
        /// Unregister a callback for device changes
        /// </summary>
        /// <param name="Handle"></param>
        /// <returns></returns>
        [DllImport("user32.dll", SetLastError=true)]
        static public extern bool UnregisterDeviceNotification(IntPtr Handle);

        /// <summary>
        /// Enumerate the interfaces of a device
        /// </summary>
        /// <param name="DeviceInfoSet"></param>
        /// <param name="DeviceInfoData"></param>
        /// <param name="InterfaceClassGuid"></param>
        /// <param name="MemberIndex"></param>
        /// <param name="DeviceInterfaceData"></param>
        /// <returns></returns>
        [DllImport("setupapi.dll", SetLastError=true)]
        static public extern int SetupDiEnumDeviceInterfaces(IntPtr DeviceInfoSet,
            int DeviceInfoData, ref System.Guid InterfaceClassGuid, int MemberIndex,
            ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData);

        /// <summary>
        /// Free an enumeration of devices
        /// </summary>
        /// <param name="DeviceInfoSet"></param>
        /// <returns></returns>
        [DllImport("setupapi.dll", SetLastError=true)]
        static public extern int SetupDiDestroyDeviceInfoList(IntPtr DeviceInfoSet);


        /// <summary>
        /// Gets a list of HidDevices
        /// </summary>
        /// <param name="ClassGuid"></param>
        /// <param name="Enumerator"></param>
        /// <param name="hwndParent"></param>
        /// <param name="Flags"></param>
        /// <returns></returns>
        [DllImport("setupapi.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern IntPtr SetupDiGetClassDevs(ref System.Guid ClassGuid,
            string Enumerator, int hwndParent, int Flags);

        /// <summary>
        /// Get the data for a device interface
        /// </summary>
        /// <param name="DeviceInfoSet"></param>
        /// <param name="DeviceInterfaceData"></param>
        /// <param name="DeviceInterfaceDetailData"></param>
        /// <param name="DeviceInterfaceDetailDataSize"></param>
        /// <param name="RequiredSize"></param>
        /// <param name="DeviceInfoData"></param>
        /// <returns></returns>
        [DllImport("setupapi.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern bool SetupDiGetDeviceInterfaceDetail(IntPtr DeviceInfoSet,
            ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData, IntPtr DeviceInterfaceDetailData,
            int DeviceInterfaceDetailDataSize, ref int RequiredSize, IntPtr DeviceInfoData);

        /// <summary>
        /// Get the GUID for HID-Devices
        /// </summary>
        /// <param name="hidGuid"></param>
        [DllImport("hid.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern void HidD_GetHidGuid(ref System.Guid hidGuid);

        /// <summary>
        /// Retrieve the attributes like vendor,product
        /// </summary>
        /// <param name="HidDeviceObject"></param>
        /// <param name="Attributes"></param>
        /// <returns></returns>
        [DllImport("hid.dll", SetLastError=true)]
        static public extern int HidD_GetAttributes(int HidDeviceObject, ref HIDD_ATTRIBUTES Attributes);

        /// <summary>
        /// Parse the caps of an interface
        /// </summary>
        /// <param name="HidDeviceObject"></param>
        /// <param name="PreparsedData"></param>
        /// <returns></returns>
        [DllImport("hid.dll", SetLastError=true)]
        static public extern bool HidD_GetPreparsedData(int HidDeviceObject, ref IntPtr PreparsedData);

        /// <summary>
        /// frees the caps of an interface
        /// </summary>
        /// <param name="PreparsedData"></param>
        /// <returns></returns>
        [DllImport("hid.dll", SetLastError=true)]
        static public extern bool HidD_FreePreparsedData(ref IntPtr PreparsedData);

        /// <summary>
        /// Gets a struct with the caps of an interface (we only need the
        /// reportsizes for the IOWarriors 
        /// </summary>
        /// <param name="PreparsedData"></param>
        /// <param name="Capabilities"></param>
        /// <returns></returns>
        [DllImport("hid.dll", SetLastError=true)]
        static public extern int HidP_GetCaps(IntPtr PreparsedData, ref HIDP_CAPS Capabilities);

        /// <summary>
        /// Get the serialnumber of the iowarrior
        /// </summary>
        /// <param name="HidDeviceObject"></param>
        /// <param name="Buffer"></param>
        /// <param name="BufferLength"></param>
        /// <returns></returns>
        [DllImport("hid.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern bool HidD_GetSerialNumberString(int HidDeviceObject, IntPtr Buffer, int BufferLength);

        /// <summary>
        /// Request the current state of the IO-Pins on the IOWarrior.
        /// This is called only once when a new device was sucessfully opend.
        /// There might be no IO-Changes firing events yet, but we always want
        /// to know what status of the IO-Pins is.
        /// </summary>
        /// <param name="HidDeviceObject"></param>
        /// <param name="Buffer"></param>
        /// <param name="BufferLength"></param>
        /// <returns></returns>
        [DllImport("hid.dll", CharSet=CharSet.Auto, SetLastError=true)]
        static public extern bool HidD_GetInputReport(int HidDeviceObject, byte [] Buffer, int BufferLength);


    }
}
