using System;
using System.Threading;
using System.Collections;
using System.Collections.Generic;
using System.Text;

namespace org.wayoda.csiow {
    /// <summary>
    /// The device with which the application communicates.
    /// </summary>
    public sealed class IOWarrior {
        /// <summary>
        /// A constant for the Product-ID of an IOWarrior24
        /// </summary>
        public const int PID_IOW24=0x1501;
        /// <summary>
        /// A constant for one of the Product-ID's of an IOWarrior24PV
        /// </summary>
        public const int PID_IOW24PV_1=0x1511;
        /// <summary>
        /// A constant for one of the Product-ID's of an IOWarrior24PV
        /// </summary>
        public const int PID_IOW24PV_2=0x1512;
        /// <summary>
        /// A constant for the Product-ID of an IOWarrior40
        /// </summary>
        public const int PID_IOW40=0x1500;
        /// <summary>
        /// A constant for the Product-ID of an IOWarrior56
        /// </summary>
        public const int PID_IOW56=0x1503;

        private const String NAME_IOW24="IOWarrior24";
        private const String NAME_IOW24PV="IOWarrior24PV";
        private const String NAME_IOW40="IOWarrior40";
        private const String NAME_IOW56="IOWarrior56";

        //A ReportListener object which encapsulates the
        //listner and the IowSpecialModes for which it registered
        private class ReportListener {
            private IowReportListener listener;
            private IowSpecialMode[] modes;
            public ReportListener(IowReportListener l, IowSpecialMode[] m) {
                listener=l;
                modes=m;
            }

            internal IowReportListener Listener {
                get {
                    return this.listener;
                }
                set {
                    this.listener=value;
                }
            }

            internal IowSpecialMode[] Modes {
                get {
                    return this.modes;
                }
                set {
                    this.modes=value;
                }
            }

            public override bool Equals(object o) {
                ReportListener rl=o as ReportListener;
                if(rl!=null && rl.Listener==this.listener)
                    return true;
                return false;
            }

            public override int GetHashCode() {
                return base.GetHashCode();
            }

        }

        //the list of listeners for reports
        private ArrayList reportListenerList=new ArrayList();

        //this is true as long as the device is plugged in
        private bool plugged;
        //an internal handle for the device
        private int handle;
        //the handle for the io-pin pipe
        private int fdIO;
        //the handle for the specialmode pipe
        private int fdSM;
        //the filename for the io-pin pipe
        private String fnIO;
        //the filename for the specialmode pipe
        private String fnSM;
        //the report-size on the io-pin pipe
        private int reportSizeIO;
        //the report-size on the specialmode pipe
        private int reportSizeSM;
        //the product-id of the device
        private int product;
        //the name of the product
        private String productName="";
        //the serial-number of the device
        private uint serial;
        //the revision of the device
        private int revision;
        //the supported SpecialModes for this device
        private IowSpecialMode[] supportedSpecialModes;
        //the latency for reports on this device
        private int deviceLatency;

        //the thread that reads from the io-pins of  this device
        private IowReadThread rtIO;
        //monitors the io-pin interface in a thread
        private Thread ioReader;
        //the thread that reads from the special-modes of  this device
        private IowReadThread rtSM;
        //monitors the special-mode interface in a thread
        private Thread smReader;
        //a private class that handles writes to the iowarrior
        private IowWriter iowWriter;

        //a mutex for creating new Handles
        private static Object handleMutex=new Object();
        //the base for creating new Handles
        private static int handleSerial=0;
        //a mutex that protects writes for arrays of IowReports
        private Object writeMutex=new Object();


        //possible states for an IOWarrior
        internal enum DeviceState {
            RUNNING,        //threads are running the device can be used
            INCOMPLETE,     //only one of the interfaces is set
            STARTUP_ERROR   //error when trying to startup the device, unusable device
        }

        //The last IowReport received from the IO-Pins. We store this
        //because it contains the current status of the IO-pins.
        private IowReport ioStatus=new IowReport();

        /// <summary>
        /// The IOWarriors are created by the IowManager, no application
        /// show therefore see an constructor
        /// </summary>
        internal IOWarrior() {
            lock(handleMutex) {
                handleSerial++;
                this.handle=handleSerial;
            }
        }

        /// <summary>
        /// Gets the Product-Id of the device.
        /// </summary>
        /// <returns>
        /// One of the constants IOWarrior.PID_IOW24, IOWarrior.PID_IOW24PV1,
        /// IOWarrior.PID_IOW24PV2,IOWarrior.PID_IOW40,IOWarrior.PID_IOW56,
        /// </returns>
        public int getProductId() {
            return product;
        }

        /// <summary>
        /// Gets the Product-Id of the IOWarrior as a String.
        /// </summary>
        /// <returns>
        /// A String with the (hex-encoded) Product-Id
        /// </returns>
        public String getProductIdString() {
            return String.Format("{0:X4}", product);
        }

        /// <summary>
        /// Gets the name of the product.
        /// </summary>
        /// <returns>
        /// Returns a product-name like "IOWarrior24, IOWarrior40 ..."
        /// </returns>
        public String getProductName() {
            return productName;
        }

        /// <summary>
        /// Gets the serial number of the IOWarrior
        /// </summary>
        /// <returns>
        /// The serialnumber for the device.
        /// </returns>
        public uint getSerialNumber() {
            return serial;
        }

        /// <summary>
        /// Gets the serialnumber of the device as a string
        /// </summary>
        /// <returns>
        /// The (hex-encoded) serialnumber of the device
        /// </returns>
        public String getSerialNumberString() {
            return String.Format("{0:X8}", serial);
        }

        /// <summary>
        /// Gets the revsision of the device
        /// </summary>
        /// <returns>
        /// The revision-number of the device
        /// </returns>
        public int getRevision() {
            return revision;
        }

        /// <summary>
        /// Gets the revision of the device as a string
        /// </summary>
        /// <returns>
        /// Gets the (Hex-encoded) revison-number of the device
        /// </returns>
        public String getRevisionString() {
            int i1, i2, i3, i4, rev;

            StringBuilder sb=new StringBuilder();
            rev=revision;
            i4=rev%16;
            rev=rev/16;
            i3=rev%16;
            rev=rev/16;
            i2=rev%16;
            rev=rev/16;
            i1=rev%16;
            sb.Append(String.Format("{0:X}", i1));
            sb.Append(".");
            sb.Append(String.Format("{0:X}", i2));
            sb.Append(".");
            sb.Append(String.Format("{0:X}", i3));
            sb.Append(".");
            sb.Append(String.Format("{0:X}", i4));
            return sb.ToString();
        }

        /// <summary>
        /// Gets the number of bytes that make up a report from the
        /// IO-Pins
        /// </summary>
        /// <returns>
        /// The number of bytes for a report from the IO-Pins
        /// </returns>
        public int getReportSizeIO() {
            return reportSizeIO;
        }

        /// <summary>
        /// Gets the number of bytes that make up a report from the
        /// SpecialModes.
        /// </summary>
        /// <returns>
        /// The number of bytes for a report from the SpecialModes
        /// </returns>
        public int getReportSizeSM() {
            return reportSizeSM;
        }

        /// <summary>
        /// Gets the minimum timespan between 2 report from an IOWarrior
        /// </summary>
        /// <returns>
        /// The minimum number of milliseconds between 2 reports from
        /// an IOWarrior.
        /// </returns>
        public int getReportLatency() {
            return deviceLatency;
        }

        /// <summary>
        /// Gets an Array of the supported SpecialModes by this device
        /// </summary>
        /// <returns>
        /// An Array with all the SpecialModes this device supports
        /// </returns>
        public IowSpecialMode[] getSupportedSpecialModes() {
            return supportedSpecialModes;
        }

        /// <summary>
        /// Tests wether this device supports a specific SpecialMode
        /// </summary>
        /// <param name="mode">
        /// The SpecialMode to be tested
        /// </param>
        /// <returns>
        /// Returns true if the device supports this special mode, false otherwise
        /// </returns>
        public bool isSupportedMode(IowSpecialMode mode) {
            bool retval=false;
            for(int i=0;i<supportedSpecialModes.Length;i++) {
                if(mode==supportedSpecialModes[i]) {
                    retval=true;
                    break;
                }
            }
            return retval;
        }

        /// <summary>
        /// Tests wether this device is still connected.
        /// </summary>
        /// <returns>
        /// Returns true if the device is connected, false if it has been unplugged.
        /// </returns>
        public bool isConnected() {
            return plugged;
        }

        /// <summary>
        /// Add a listener for reports from this device. 
        /// </summary>
        /// <param name="listener">
        /// An instance of a class that implements the interface IowReportListener. This
        /// listener will be notified when new reports are read from the IOWarrior.
        /// If the listener was already added by a previous call to addReportListener
        /// only the existing listeners SpecialModes will be updated.
        /// </param>
        /// <param name="modes">
        /// An Array of SpecialModes that the listener is interested in. If the Array
        /// is null or empty, the listener will be notified for every type of report from
        /// the device.
        /// </param>
        public void addReportListener(IowReportListener listener, IowSpecialMode[] modes) {
            ReportListener l=new ReportListener(listener, modes);
            lock(reportListenerList) {
                for(int i=0;i<reportListenerList.Count;i++) {
                    ReportListener rl=(ReportListener)reportListenerList[i];
                    if(rl.Listener.Equals(l)) {
                        rl.Modes=modes;
                        return;
                    }
                }
                reportListenerList.Add(l);
            }
        }

        /// <summary>
        /// Remove a listner for reports from this device.
        /// </summary>
        /// <param name="listener">
        /// The listener that is not interested in reports any more.
        /// </param>
        public void removeReportListener(IowReportListener listener) {
            lock(reportListenerList) {
                for(int i=0;i<reportListenerList.Count;) {
                    ReportListener l=(ReportListener)reportListenerList[i];
                    if(l.Listener==listener)
                        reportListenerList.RemoveAt(i);
                    else
                        i++;
                }
            }
        }

        public IowReport getIOStatus() {
            lock(ioStatus) {
                return new IowReport(ioStatus);
            }
        }

        /// <summary>
        /// update all listeners with a new report for the independent reader threads
        /// </summary>
        /// <param name="rep">
        /// The report that was read
        /// </param>
        internal void reportUpdate(IowReport rep) {
            lock(ioStatus) {
                if(rep[0]==0x00) {
                    ioStatus=new IowReport(rep);
                }
            }
            fireReportEvent(rep);
        }

        /// <summary>
        /// Send a new report to all listeners that registered for the
        /// SpecialMode of this report.
        /// </summary>
        /// <param name="rep">
        /// The report that was read from the device
        /// </param>
        private void fireReportEvent(IowReport rep) {
            lock(reportListenerList) {
                for(int i=0;i<reportListenerList.Count;i++) {
                    bool wantUpdate=false;
                    ReportListener l=(ReportListener)reportListenerList[i];
                    IowSpecialMode[] m=l.Modes;
                    if(m!=null && m.Length>0) {
                        for(int j=0;j<m.Length;j++) {
                            if(rep.getSpecialMode()==m[j]) {
                                wantUpdate=true;
                                break;
                            }
                        }
                    }
                    else
                        wantUpdate=true;
                    if(wantUpdate) {
                        l.Listener.reportUpdate(rep, this);
                    }
                }
            }
        }

        /// <summary>
        /// Tell all listeners the device is unplugged
        /// </summary>
        private void fireUnplugEvent() {
            lock(reportListenerList) {
                for(int i=0;i<reportListenerList.Count;i++) {
                    ReportListener l=(ReportListener)reportListenerList[i];
                    l.Listener.deviceRemoved(this);
                }
            }
        }

        /// <summary>
        /// Writes a single IowReport to the device
        /// </summary>
        /// <param name="rep">
        /// The data to be written
        /// </param>
        /// <returns>If the data was successfully written to the device we return
        /// IowError.OK. Otherwise the write failed an the type of IowError could be
        /// inspected for the reason of the failure.
        /// </returns>
        public IowError write(IowReport rep) {
            lock(writeMutex) {
                if(rep==null)
                    return new IowError(IowError.Type.WRITE_NULL, this,
                        "IowReport was null in write", 0);
                if(!isSupportedMode(IowSpecialMode.getByReportId(rep.getReportId())))
                    return new IowError(IowError.Type.WRITE_UNSUPPORTED_MODE, this,
                        "This device does not support the SpecialMode found in the ReportId",
                        0);
                if(!isConnected())
                    return new IowError(IowError.Type.UNPLUGGED, this,
                        "This device is unplugged",
                        0);
                return iowWriter.write(rep);
            }
        }

        /// <summary>
        /// Writes an array of reports to the device. This call is thread-safe,
        /// as it locks the write()-call for other threads.
        /// </summary>
        /// <param name="reps">An array of IowReports</param>
        /// <returns>
        /// IowError.OK if the write for all reports was 
        /// succesfull. Otherwise the error for the report is returned
        /// that we failed to write. If you encounter an error you can not
        /// tell which of the reports caused it, and how many of the
        /// reports were written sucessfully.
        /// </returns>
        public IowError write(IowReport[] reps) {
            IowError retval=IowError.OK;
            if(reps==null || reps.Length==0)
                return new IowError(IowError.Type.WRITE_NULL, this,
                     "IowReport Array was null or empty in write", 0);
            lock(writeMutex) {
                for(int i=0;i<reps.Length;i++) {
                    retval=write(reps[i]);
                    if(retval!=IowError.OK)
                        break;
                }
                return retval;
            }
        }

        /// <summary>
        /// Called from the IowManager when the device was reported unplugged.
        /// </summary>
        internal void setUnplugged() {
            plugged=false;
            if(rtIO!=null)
                rtIO.shutdown();
            if(rtSM!=null)
                rtSM.shutdown();
            if(ioReader!=null && ioReader.IsAlive)
                ioReader.Join();
            if(smReader!=null && smReader.IsAlive)
                smReader.Join();
            if(iowWriter!=null)
                iowWriter.shutdown();
            HidDevice.CloseHandle(fdIO);
            HidDevice.CloseHandle(fdSM);
            fdIO=0;
            fdSM=0;
            fnIO="";
            fnSM="";
            fireUnplugEvent();
        }

        /// <summary>
        /// Gets the filename of the io-interface
        /// </summary>
        /// <returns>
        /// The filename of the io-Interface
        /// </returns>
        internal String getFilenameIO() {
            return fnIO;
        }

        /// <summary>
        /// Gets the handle of the io-interface
        /// </summary>
        /// <returns>
        /// the handle for the io-interface
        /// </returns>
        internal int getFileDescriptorIO() {
            return fdIO;
        }

        /// <summary>
        /// Gets the filename of the specialmode-interface
        /// </summary>
        /// <returns>
        /// The filename for the specialmode-interface
        /// </returns>
        internal String getFilenameSM() {
            return fnSM;
        }

        /// <summary>
        /// Gets the handle of the specialmode-interface
        /// </summary>
        /// <returns>
        /// the handle of the specialmode-interface
        /// </returns>
        internal int getFileDescriptorSM() {
            return fdSM;
        }
        /// <summary>
        /// Tests wether one of the filenames for this device matches the one
        /// passed in.
        /// </summary>
        /// <param name="fn">
        /// The filename to be tested.
        /// </param>
        /// <returns>
        /// Returns true, if either the filename for the io-interface
        /// or the specialmode-interface matches argument fn.
        /// </returns>
        internal bool fileNameMatch(String fn) {
            return (this.fnIO.Equals(fn) || this.fnSM.Equals(fn));
        }

        /// <summary>
        /// Tests wether this is the device with a specific product-id 
        /// and serial-number.
        /// </summary>
        /// <param name="product">The product-id to be tested</param>
        /// <param name="serial">The serial-number to be tested</param>
        /// <returns>Returns true, if this device matches the product-id 
        /// and the serial-number agruments.
        /// </returns>
        internal bool matchDevice(int product, uint serial) {
            return (this.product==product && this.serial==serial);
        }

        /// <summary>
        /// Called from the IowManager when a new interface has been
        /// detetced in the HID-Collection.
        /// </summary>
        /// <param name="fn">The filename of the interface</param>
        /// <param name="fd">The handle of the interface</param>
        /// <param name="product">The product-Id of the device</param>
        /// <param name="serial">The serial-number of the device</param>
        /// <param name="revision">The revision of the device</param>
        /// <param name="ifNum">the type of interface reported,
        /// 0==IO_Pin; 1==SpecuialModes
        /// </param>
        /// <returns>
        /// Returns the new DeviceState if this IOWarrior:
        /// DeviceState.INCOMPLETE we have only one of the interfaces yet
        /// DeviceState.STARTUP_ERROR there was an error in starting the device
        /// it is unusable.
        /// DeviceState.RUNNING the IOWarrior is setup and can be reported to
        /// the device-listeners.
        /// </returns>
        internal DeviceState addInterface(String fn, int fd, int product,
            uint serial, int revision, int ifNum) {
            if(ifNum==0) {
                fnIO=fn;
                fdIO=fd;
            }
            else {
                fnSM=fn;
                fdSM=fd;
            }
            if(this.product==0) {
                //this is the first interface for the IOWarrior
                //we need to fill in the device caps
                this.product=product;
                this.serial=serial;
                this.revision=revision;
                switch(product) {
                    case PID_IOW24:
                        reportSizeIO=3;
                        reportSizeSM=8;
                        deviceLatency=8;
                        productName=NAME_IOW24;
                        if(revision>=0x1021) {
                            IowSpecialMode[] sm24={IowSpecialMode.IO, IowSpecialMode.IIC, 
                                IowSpecialMode.LCD, IowSpecialMode.SPI,
                                IowSpecialMode.RC5, IowSpecialMode.LED_MATRIX, 
                                IowSpecialMode.PIN_STATUS};
                            supportedSpecialModes=sm24;
                        }
                        else {
                            IowSpecialMode[] sm24={IowSpecialMode.IO, IowSpecialMode.IIC, 
                                IowSpecialMode.LCD, IowSpecialMode.SPI, 
                                IowSpecialMode.RC5, IowSpecialMode.PIN_STATUS};
                            supportedSpecialModes=sm24;
                        }
                        break;
                    case PID_IOW24PV_1:
                    case PID_IOW24PV_2:
                        reportSizeIO=3;
                        reportSizeSM=8;
                        deviceLatency=8;
                        productName=NAME_IOW24PV;
                        IowSpecialMode[] sm24pv={IowSpecialMode.IO, IowSpecialMode.IIC, 
                                IowSpecialMode.SPI,IowSpecialMode.POWER_STATUS};
                        supportedSpecialModes=sm24pv;
                        break;
                    case PID_IOW40:
                        reportSizeIO=5;
                        reportSizeSM=8;
                        deviceLatency=8;
                        productName=NAME_IOW40;
                        if(revision>=0x1021) {
                            IowSpecialMode[] sm40={IowSpecialMode.IO, IowSpecialMode.IIC, 
                                IowSpecialMode.LCD, IowSpecialMode.LED_MATRIX, 
                                IowSpecialMode.SWITCH_MATRIX, IowSpecialMode.PIN_STATUS};
                            supportedSpecialModes=sm40;
                        }
                        else {
                            IowSpecialMode[] sm40={ IowSpecialMode.IO, IowSpecialMode.IIC, 
                                IowSpecialMode.LCD, IowSpecialMode.PIN_STATUS };
                            supportedSpecialModes=sm40;

                        }
                        break;
                    case PID_IOW56:
                        reportSizeIO=8;
                        reportSizeSM=64;
                        deviceLatency=1;
                        productName=NAME_IOW56;
                        IowSpecialMode[] sm56={IowSpecialMode.IO, IowSpecialMode.IIC, 
                            IowSpecialMode.LCD, IowSpecialMode.SPI,
                            IowSpecialMode.SWITCH_MATRIX, IowSpecialMode.LED_MATRIX, 
                            IowSpecialMode.PIN_STATUS};
                        supportedSpecialModes=sm56;
                        break;
                }
            }
            plugged=true;
            if(fdIO!=0 && fdSM!=0) {
                rtIO=new IowReadThread(this, IowReadThread.INTERFACE_IO);
                rtSM=new IowReadThread(this, IowReadThread.INTERFACE_SM);
                ThreadStart ts=new ThreadStart(rtIO.run);
                ioReader=new Thread(ts);
                try {
                    ioReader.Start();
                }
                catch(Exception e) {
                    //this here is a really stupid construct
                    //for supressing a warning from the compiler about
                    //an unused variable
                    if(e!=null)
                        e=null;
                    //End of stupid construct
                    return DeviceState.STARTUP_ERROR;
                }
                ts=new ThreadStart(rtSM.run);
                smReader=new Thread(ts);
                try {
                    smReader.Start();
                }
                catch(Exception e) {
                    //this here is a really stupid construct
                    //for supressing a warning from the compiler about
                    //an unused variable
                    if(e!=null)
                        e=null;
                    //End of stupid construct
                    ioReader.Abort();
                    ioReader.Join();
                    return DeviceState.STARTUP_ERROR;
                }
                iowWriter=new IowWriter(this);
                if(!iowWriter.isSetup()) {
                    //setup for overlapped write failed
                    ioReader.Abort();
                    smReader.Abort();
                    ioReader.Join();
                    smReader.Join();
                    return DeviceState.STARTUP_ERROR;
                }
                return DeviceState.RUNNING;
            }
            return DeviceState.INCOMPLETE;
        }

        /// <summary>
        /// Gets a string-representation of the device.
        /// </summary>
        /// <returns>
        /// Returns a String with the product-name, serial, revision.
        /// If the device is unplugged the string "[unplugged]" will be added.
        /// </returns>
        public override string ToString() {
            StringBuilder sb=new StringBuilder();
            sb.Append(getProductName());
            sb.Append(" ");
            sb.Append(getSerialNumberString());
            sb.Append(" ");
            sb.Append(getRevisionString());
            if(!plugged)
                sb.Append(" [unplugged]");
            return sb.ToString();
        }

    }

}
