using System;
using System.Collections.Generic;
using System.Text;

namespace org.wayoda.csiow {
    /// <summary>
    /// Errors that are returned from operation in the library.
    /// </summary>
    public sealed class IowError {
        /// <summary>
        /// The type of error that was encountered. Type is one of the constants
        /// NO_ERROR - The operation completed without error
        /// LIB_ERROR - The was an unrecoverable error, when trying to
        ///             setup the library.
        /// WRITE_NULL - You tried to write an null-object to an IOWarrior
        /// WRITE_TIMEOUT - we encountered a 5 second timeout when writing
        ///                 to an IOWarrior
        /// WRITE_UNSUPPORTED_MODE - You tried to send an IowReport with a report-id
        ///                          that is not supported by the IOWarrior to which
        ///                          it was sent.
        /// WRITE_FAILED - There was an error in the unmanaged code for writing
        ///                data. 
        /// UNPLUGGED - You tried to initiate an operation on an IOWarrior that
        ///             is already unplugged.
        /// </summary>
        public enum Type {
            NO_ERROR,               //no error
            LIB_ERROR,              //error in library
            WRITE_NULL,             //tried to write a null report
            WRITE_TIMEOUT,          //write failed with timeout
            WRITE_UNSUPPORTED_MODE, //tried to write report with unsupported specialmode
            WRITE_FAILED,           //write failed unspecified
            UNPLUGGED               //device is unplugged
        }

        private Type t;
        private IOWarrior iow;
        private String msg;
        private int win32Code;

        /// <summary>
        /// A typed constant for a successfully completed operation.
        /// </summary>
        public static readonly IowError OK=new IowError(Type.NO_ERROR,
            null, "Operation completed without error", 0);

        /// <summary>
        /// Creates a new IowError.
        /// </summary>
        /// <param name="t">The type of error</param>
        /// <param name="iow">The device on which the error was detected.
        /// this is null for errors of type LIB_ERROR.
        /// </param>
        /// <param name="msg">A descrption of the error</param>
        /// <param name="win32Code">If the error was detected in unmanged code,
        /// this will be the code from Marshal.GetLastWin32Error(). 0 otherwise.
        /// </param>
        internal IowError(Type t, IOWarrior iow, String msg, int win32Code) {
            this.t=t;
            this.iow=iow;
            this.msg=msg;
            this.win32Code=win32Code;
        }

        /// <summary>
        /// Gets the type of error that was detected.
        /// </summary>
        /// <returns>
        /// One of the types defined in IowError.Type.
        /// </returns>
        public Type getType() {
            return t;
        }

        /// <summary>
        /// Gets the IOWarrior on which the error was detected.
        /// </summary>
        /// <returns>
        /// The IOWarrior on which the error was deteced. This will be null
        /// for an error with type IowError.LIB_ERROR.
        /// </returns>
        public IOWarrior getDevice() {
            return iow;
        }

        /// <summary>
        /// Gets a description of the errro that was detected
        /// </summary>
        /// <returns>
        /// A short description of the error.
        /// </returns>
        public String getMessage() {
            return msg;
        }

        /// <summary>
        /// If the error was detected in unmanaged code this will return
        /// the Win32-constant for this error.
        /// </summary>
        /// <returns> The win32-code for this error if the error was detected in
        /// unmanaged code. 0 otherwise.
        /// </returns>
        public int getWin32Code() {
            return win32Code;
        }

        /// <summary>
        /// A string with a detailed description of the error.
        /// </summary>
        /// <returns>A description of the error</returns>
        public override string ToString() {
            StringBuilder sb=new StringBuilder();
            sb.Append("IowError : ");
            switch(t) {
                case Type.NO_ERROR:
                    sb.Append("OK ");
                    if(this.iow!=null) {
                        sb.Append(this.iow);
                        sb.Append(" ");
                    }
                    sb.Append(this.msg);
                    break;
                case Type.LIB_ERROR:
                    sb.Append("Library failure ");
                    sb.Append(this.msg);
                    break;
                case Type.UNPLUGGED:
                    sb.Append("Device unplugged ");
                    sb.Append(this.iow);
                    sb.Append(" ");
                    sb.Append(this.msg);
                    break;
                case Type.WRITE_FAILED:
                case Type.WRITE_NULL:
                case Type.WRITE_TIMEOUT:
                case Type.WRITE_UNSUPPORTED_MODE:
                    sb.Append("Device Write failure ");
                    sb.Append(this.iow);
                    break;
            }
            if(this.win32Code!=0) {
                sb.Append(" (Win32NativeCode=");
                sb.Append(this.win32Code);
                sb.Append(")");
            }
            return sb.ToString();
        }

    }
}
