using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace org.wayoda.csiow {
    /// <summary>
    /// A unique handle for an IOWarrior. Unique in this case
    /// means that IOWarriors that are unplugged and plugged in again
    /// will have different handles even though its the same device
    /// It is created by the IowDeviceManger. For applications that
    /// use our class-library it is just an immutable object.
    /// </summary>
    internal class IowHandle {
        private static int handleSerial=0;
        private int id;
        private int product;
        private uint serial;
        private int revision;
        private String fnIO;
        private int fdIO;
        private String fnSM;
        private int fdSM;
        private bool unplugged;
        private static Object serialMutex=new Object();

        private static int getHandleSerial() {
            lock(serialMutex) {
                handleSerial++;
            }
            return handleSerial;
        }
            
        /// <summary>
        /// The internal constructor is called by the IowDevicemanager
        /// when an new IOWarrior is reported.
        /// </summary>
        /// <param name="id">
        /// the internal id for the IowDeviceManager 
        /// </param>
        internal IowHandle() {
            this.id=getHandleSerial();
        }
        
        /// <summary>
        /// IowHandles are used as a key in a hashtable so we need to implement
        /// this method.
        /// </summary>
        /// <param name="o">
        /// The IowHandle to be compared to this one.
        /// </param>
        /// <returns>
        /// bool True if the argument o is an IowHandle and 
        /// the internal id matches to this one
        /// </returns>
        public override bool Equals(Object o) {
            if(o!=null && o is IowHandle) {
                IowHandle h=(IowHandle)o;
                if(h.id==id)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// IowHandles are used as a key in a hashtable so we need to implement
        /// this method.
        /// </summary>
        /// <returns>
        /// int the internal id for this IowHandle (which is unique)
        /// </returns>
        public override int GetHashCode() {
            return id;
        }

        /// <summary>
        /// Gets the internal id for this IowHandle
        /// </summary>
        /// <returns>
        /// int the internal id for the handle
        /// </returns>
        internal int getHandleValue() {
            return id;
        }

        internal bool fileNameMatchIO(String fn) {
            return String.Compare(fnIO, fn)==0;
        }

        internal bool fileNameMatchSM(String fn) {
            return String.Compare(fnSM, fn)==0;
        }

        internal bool fileNameMatch(String fn) {
            return (fileNameMatchIO(fn) || fileNameMatchSM(fn)); 
        }

        internal bool isUnplugged() {
            return unplugged;
        }

        internal void setUnplugged() {
            unplugged=true;
        }
        
        internal int getFileHandleIO() {
            return fdIO;
        }

        internal int getFileHandleSM() {
            return fdSM;
        }

        internal int getProduct() {
            return product;
        }

        internal uint getSerial() {
            return serial;
        }

        internal int getRevision() {
            return revision;
        }

        internal void setInterface(String fn, int fd, int product, uint serial, int revision, int ifNum) {
            if(ifNum==0) {
                fnIO=fn;
                fdIO=fd;
            }
            else {
                fnSM=fn;
                fdSM=fd;
            }
            this.product=product;
            this.serial=serial;
            this.revision=revision;
        }

        internal bool matchDevice(int product, uint serial) {
            return (this.product==product && this.serial==serial);
        }

        public override string ToString() {
            StringBuilder sb=new StringBuilder();
            sb.Append("handle=");
            sb.Append(id);
            sb.Append(" product=");
            sb.Append(product.ToString("x"));
            sb.Append(" serial=");
            sb.Append(serial.ToString("x"));
            sb.Append(" revision=");
            sb.Append(revision.ToString("x"));
            sb.Append("\nfnIO=");
            sb.Append(fnIO);
            sb.Append("\nfnSM=");
            sb.Append(fnSM);
            return sb.ToString();
        }
    }
}
