using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;

namespace org.wayoda.csiow {
    /// <summary>
    /// Manages all the IOWarriors found on the system. This class implements
    /// only static methods and has only a private constructor. So there will never
    /// be an instance of this class in the project.
    /// </summary>
    public sealed class IowManager {
        //here we store all devices 
        private static ArrayList deviceList=new ArrayList();
        //the delegate for device change events reported by the MessageCatcher
        private static DeviceChangeEventHandler devChangeDelegate=new DeviceChangeEventHandler(onDeviceChange);
        //a simple invisible window that listens for WM_DEVICECHANGE messages
        private static MessageCatcher mc=new MessageCatcher(devChangeDelegate);
        //the handle which we register for device changes
        private static IntPtr deviceNotifyHandle=new IntPtr();
        //a mutex that serializes devicechange callbacks
        private static Object deviceChangeMutex=new Object();
        //a mutex that protects the open() and close() calls from multiple threads
        private static Object ocMutex=new Object();
        //this will be set to true on open and reset to false at the end of close
        //prevents that open is called more than once
        private static bool isopen=false;
        //The object that will be notified about device add/removal
        private static IowDeviceChangeListener devListener=null;
        //the filter for devices that we accept
        private static IowDeviceFilter dfilter;

        /// <summary>
        /// There will be no instances of this class, everything is static in here
        /// </summary>
        private IowManager() {
        }

        /// <summary>
        /// Opens the manager for IOWarriors. It will start by scanning the system for all
        /// currently attached IOWarriors. If any are found they are reported to the
        /// IowDeviceChangeListener registered with the open call. Then it will start
        /// to monitor the system for plug/unplug-events caused by IOWarriors. All events
        /// will be reported to the listener until IowManager.close() is called.
        /// Any attempt to call IowManager.open() twice without a previous call to
        /// IowManager.close() will throw an InvalidOperationException.
        /// </summary>
        /// <param name="listener">
        /// An instance of a class that implements the IowDeviceChangeListener interface.
        /// If listener is null an Exception of type ArgumentNullExcption is thrown.
        /// </param>
        /// <returns>
        /// An IowError. If the mamager is up and running we return the typed
        /// constant IowError.OK, otherwise the errors message could be inspected
        /// for more information about what went wrong. 
        /// If the open() call succeeded (IowError.OK returned) you have to call
        /// IowManager.close() when youre done with the library to free up all resources
        /// held by the devices. There are Threads running for each device so simply
        /// closing your app will NOT WORK!
        /// </returns>
        public static IowError open(IowDeviceChangeListener listener) {
            return open(listener, new IowDeviceFilter());
        }

        
        /// <summary>
        /// Opens the manager for IOWarriors. It will start by scanning the system for all
        /// currently attached IOWarriors. If any are found they are reported to the
        /// IowDeviceChangeListener registered with the open call. Then it will start
        /// to monitor the system for plug/unplug-events caused by IOWarriors. All events
        /// will be reported to the listener until IowManager.close() is called.
        /// Any attempt to call IowManager.open() twice without a previous call to
        /// IowManager.close() will throw an InvalidOperationException.
        /// </summary>
        /// <param name="listener">
        /// An instance of a class that implements the IowDeviceChangeListener interface.
        /// If listener is null an Exception of type ArgumentNullExcption is thrown.
        /// </param>
        /// <param name="deviceFilter">
        /// A filter for devices to be accepted by the library. If filter is null
        /// we accept every IOWarrior. The filter must be setup completly before
        /// you call this method. 
        /// </param>
        /// <returns>
        /// An IowError. If the mamager is up and running we return The typed
        /// constant IowError.OK, otherwise the errors message could be inspected
        /// for more information about what went wrong. 
        /// If the open() call succeeded (IowError.OK returned) you have to call
        /// IowManager.close() when youre done with the library to free up all resources
        /// held by the devices. There are Threads running for each device so simply
        /// closing your app will NOT WORK!
        /// </returns>
        public static IowError open(IowDeviceChangeListener listener,
            IowDeviceFilter deviceFilter) {
            if(System.Environment.OSVersion.Platform!=System.PlatformID.Win32NT
                && System.Environment.OSVersion.Version.Major<5
                && System.Environment.OSVersion.Version.Minor<1)
                throw new System.PlatformNotSupportedException("Only Windows XP or newer is supported");
            lock(ocMutex) {
                if(isopen==true) {
                    //we were called twice, and we are ANGRY NOW!
                    throw new System.InvalidOperationException("IowManager cannot be opened twice");
                }
                else
                    isopen=true;
                if(listener==null) {
                    isopen=false;
                    throw new System.ArgumentNullException("You have to supply a listener for device changes");
                }
            }
            devListener=listener;
            if(deviceFilter==null)
                deviceFilter=new IowDeviceFilter();
            dfilter=deviceFilter;
            try {
                //Get a list for all hid-device filenames
                ArrayList hidPathList=getHidDevicePathList();
                //Check wether there are IOWarriors in this list
                findIows(hidPathList);
                //now register for asynch device changes
                addDeviceChangeCatcher();
            }
            catch(IowManagerException e) {
                //we probably fired some device-add messages already
                //we gotta take them back
                lock(deviceList) {
                    while(deviceList.Count>0) {
                        IOWarrior iow=(IOWarrior)deviceList[0];
                        iow.setUnplugged();
                        fireDeviceChange(MessageCatcher.DEVICE_REMOVED, iow); 
                        devListener.deviceRemoved(iow);
                        deviceList.RemoveAt(0);
                    }
                }
                devListener=null;
                isopen=false;
                return e.getError();
            }
            return IowError.OK;
        }

        /// <summary>
        /// Tests wether the library is open.
        /// </summary>
        /// <returns>True, if a call to IowManager.open() was successful and
        /// a final call to IowManager.close() is required</returns>
        public static bool isOpen() {
            return isopen;
        }

        /// <summary>
        /// Close the IowManager. All IOWarriors will be reported as unplugged
        /// to the IowDeviceChangeListener, and the IowReportListeners attached to 
        /// a specific IOWarrior.
        /// </summary>
        public static void close() {
            if(isopen==false)
                return;
            removeDeviceChangeCatcher();
            //mark all devices unplugged (notifies all listeners) and forget about them
            lock(deviceList) {
                while(deviceList.Count>0) {
                    IOWarrior iow=(IOWarrior)deviceList[0];
                    iow.setUnplugged();
                    devListener.deviceRemoved(iow);
                    deviceList.RemoveAt(0);
                }
            }
            devListener=null;
            lock(ocMutex) {
                isopen=false;
            }
        }


        /// <summary>
        /// Send a device change event to all listeners. Each device change is reported in its own
        /// thread.
        /// </summary>
        /// <param name="action"></param>
        /// <param name="iow"></param>
        private static void fireDeviceChange(int action, IOWarrior iow) {
            DeviceNotifyThread nt=new DeviceNotifyThread(devListener, action, iow);
            Thread t=new Thread(nt.run);
            t.Start();
        }

        /// <summary>
        /// This is called from the MessageCatcher when it detects a device change
        /// </summary>
        /// <param name="changeAction">
        /// One of the constants MessageCatcher.DEVICE_REMOVED or
        /// MessageCatcher.DEVICE_ADD
        /// </param>
        private static void onDeviceChange(int changeAction) {
            lock(deviceChangeMutex) {
                ArrayList hidPathList=null;
                //get list of filenames found on the hid-collection
                try {
                    hidPathList=getHidDevicePathList();
                }
                catch(IowManagerException e) {
                    //this here is a really stupid construct
                    //for supressing a warning from the compiler about
                    //an unused variable
                    if(e!=null)
                        e=null;
                    //End of stupid construct

                    //well what can we do ?
                    // what about an ErrorHandler?
                    return;
                }
                //get the list of filenames for all managed IOWarriors
                ArrayList devicePathList=getDevicePathList();
                if(changeAction==MessageCatcher.DEVICE_ADD) {
                    //remove all attached devices from the hid-collection
                    for(int i=0;i<devicePathList.Count;i++) {
                        String fn=(String)devicePathList[i];
                        for(int j=0;j<hidPathList.Count;) {
                            String hidFn=(String)hidPathList[j];
                            if(fn.Equals(hidFn))
                                hidPathList.RemoveAt(j);
                            else
                                j++;
                        }
                    }
                    //whats left here could be new IOWarriors
                    //now open them and notify our listener for
                    //device changes
                    findIows(hidPathList);
                }
                else if(changeAction==MessageCatcher.DEVICE_REMOVED) {
                    //remove all entrys from the hid-collection
                    //that show up in the list of managed IOWarriors
                    for(int i=0;i<hidPathList.Count;i++) {
                        String hidFn=(String)hidPathList[i];
                        for(int j=0;j<devicePathList.Count;) {
                            String fn=(String)devicePathList[j];
                            if(fn.Equals(hidFn))
                                devicePathList.RemoveAt(j);
                            else
                                j++;
                        }
                    }
                    //whats left here are the filenames that do not
                    //show up in the hid-collection any more
                    //So these are IOWarriors that have been unplugged
                    for(int i=0;i<devicePathList.Count;i++) {
                        String fn=(String)devicePathList[i];
                        for(int j=0;j<deviceList.Count;j++) {
                            IOWarrior iow=(IOWarrior)deviceList[j];
                            if(iow.fileNameMatch(fn)) {
                                iow.setUnplugged();
                                fireDeviceChange(MessageCatcher.DEVICE_REMOVED, iow);
                                deviceList.RemoveAt(j);
                                break;
                            }
                        }
                    }
                }
            }
        }
        
        /// <summary>
        /// Gets a list of all the filenames found in currently
        /// managed IOWarriors.
        /// </summary>
        /// <returns>
        /// An ArrayList that contains the filenames of each IOWarrior
        /// currently managed. If no IOWarriors are currently managed
        /// we return an empty list.
        /// </returns>
        private static ArrayList getDevicePathList() {
            ArrayList al=new ArrayList();
            for(int i=0;i<deviceList.Count;i++) {
                IOWarrior iow=(IOWarrior)deviceList[i];
                al.Add(iow.getFilenameIO());
                al.Add(iow.getFilenameSM());
            }
            return al;
        }

        /// <summary>
        /// Gets a list of filenames for all currently attached HID-Devices
        /// </summary>
        /// <returns>
        /// An ArrayList with the filenames of all hid-devices found by 
        /// enuration the HID-Collection. If no devices are found on the collection
        /// we return an empty list.
        /// </returns>
        private static ArrayList getHidDevicePathList() {
            int result=0;
            bool bresult=false;
            int devCount=0;
            bool hasMoreDevices=true;

            //a list of all the filenames found when enumerating the IOWarriors
            ArrayList hidPathList=new ArrayList();
            HidDevice.SP_DEVICE_INTERFACE_DATA iData=new HidDevice.SP_DEVICE_INTERFACE_DATA();
            iData.cbSize=28;
            HidDevice.SP_DEVICE_INTERFACE_DETAIL_DATA iDetailData=new HidDevice.SP_DEVICE_INTERFACE_DETAIL_DATA();

            IntPtr devList;
            //get the Guid for HID-Devs
            System.Guid g=new System.Guid();
            HidDevice.HidD_GetHidGuid(ref g);
            //get a list of plugged devices
            devList=HidDevice.SetupDiGetClassDevs(ref g, null, 0, HidDevice.DIGCF_PRESENT | HidDevice.DIGCF_DEVICEINTERFACE);
            if(devList.ToInt32()==HidDevice.INVALID_HANDLE_VALUE)
                throw new IowManagerException(new IowError(IowError.Type.LIB_ERROR, null,
                    "Unable to retrieve the list of HID-Devices", Marshal.GetLastWin32Error()));
            do {
                result=HidDevice.SetupDiEnumDeviceInterfaces(devList, 0, ref g, devCount, ref iData);
                if(result==0) {
                    //no more HID-Devices, we're done
                    hasMoreDevices=false;
                }
                else {
                    //now get information about the device
                    int bufferSize=-1;
                    bresult=HidDevice.SetupDiGetDeviceInterfaceDetail(devList,
                        ref iData, IntPtr.Zero, 0, ref bufferSize, IntPtr.Zero);
                    if(bufferSize==-1) {
                        //we have a problem with this interface, skip it
                        continue;
                    }
                    iDetailData.cbSize=Marshal.SizeOf(iDetailData);
                    IntPtr iDetailDataBuffer = Marshal.AllocHGlobal(bufferSize);
                    Marshal.WriteInt32(iDetailDataBuffer, 4 + Marshal.SystemDefaultCharSize);
                    bresult=HidDevice.SetupDiGetDeviceInterfaceDetail(devList,
                        ref iData, iDetailDataBuffer, bufferSize, ref bufferSize, IntPtr.Zero);
                    if(bresult==false) {
                        //we have a problem with this interface, skip it
                        continue;
                    }
                    //now try to get the pathname for the device
                    IntPtr devPathNameBuffer = new IntPtr(iDetailDataBuffer.ToInt32() + 4);

                    // Get the String containing the devicePathName.
                    String devPathName = Marshal.PtrToStringAuto(devPathNameBuffer);
                    if(devPathName==null) {
                        //we have a problem with this interface, skip it
                        continue;
                    }
                    hidPathList.Add(devPathName);
                    // Free the memory allocated previously by AllocHGlobal.
                    Marshal.FreeHGlobal(iDetailDataBuffer);
                    //now iterate into next loop
                    devCount++;
                }
            } while(hasMoreDevices);
            //free resources from enuerating the hid-devices
            HidDevice.SetupDiDestroyDeviceInfoList(devList);
            return hidPathList;
        }

        /// <summary>
        /// Find and create new IOWarriors from a list of filenames. 
        /// </summary>
        /// <param name="hidPathList">
        /// A list of filenames which are possibly the interfaces of an IOWarrior
        /// </param>
        private static void findIows(ArrayList hidPathList) {
            int result;
            bool bresult;
            int fd;
            int product;
            int revision;
            uint serialNumber;
            int interfaceNumber=0;
            HidDevice.SECURITY_ATTRIBUTES sec;
            HidDevice.HIDD_ATTRIBUTES hidAttributes;
            HidDevice.HIDP_CAPS hidCaps;

            if(hidPathList==null || hidPathList.Count==0)
                return;
            sec=new HidDevice.SECURITY_ATTRIBUTES();
            sec.lpSecurityDescriptor = 0;
            sec.bInheritHandle = System.Convert.ToInt32(true);
            sec.nLength = Marshal.SizeOf(sec);
            for(int i=0;i<hidPathList.Count;i++) {
                fd=HidDevice.CreateFile((String)hidPathList[i],
                    HidDevice.GENERIC_READ | HidDevice.GENERIC_WRITE,
                    HidDevice.FILE_SHARE_READ | HidDevice.FILE_SHARE_WRITE,
                    ref sec, HidDevice.OPEN_EXISTING, HidDevice.FILE_FLAG_OVERLAPPED, 0);
                if(fd==HidDevice.INVALID_HANDLE_VALUE) {
                    //could not open the file
                    continue;
                }
                hidAttributes=new HidDevice.HIDD_ATTRIBUTES();
                hidAttributes.Size=Marshal.SizeOf(hidAttributes);
                hidCaps=new HidDevice.HIDP_CAPS();
                result = HidDevice.HidD_GetAttributes(fd, ref hidAttributes);
                if(result==0) {
                    //getting hid-attributes failed
                    HidDevice.CloseHandle(fd);
                    continue;
                }
                if(hidAttributes.VendorID!=0x07c0) {
                    //not a codemercs device
                    HidDevice.CloseHandle(fd);
                    continue;
                }
                product=hidAttributes.ProductID;
                revision=hidAttributes.VersionNumber;
                if(!(product==IOWarrior.PID_IOW24
                  ||product==IOWarrior.PID_IOW24PV_1
                  ||product==IOWarrior.PID_IOW24PV_2
                  ||product==IOWarrior.PID_IOW40
                  ||product==IOWarrior.PID_IOW56)) {
                    //not a supported product
                    HidDevice.CloseHandle(fd);
                    continue;
                }
                IntPtr ppDataPointer = new IntPtr();
                bresult=HidDevice.HidD_GetPreparsedData(fd, ref ppDataPointer);
                if(!bresult) {
                    //unable to get preparsed data
                    HidDevice.CloseHandle(fd);
                    continue;
                }
                result=HidDevice.HidP_GetCaps(ppDataPointer, ref hidCaps);
                if(result==0) {
                    //unable to get device caps
                    HidDevice.CloseHandle(fd);
                    continue;
                }
                HidDevice.HidD_FreePreparsedData(ref ppDataPointer);
                switch(product) {
                    // check for the interface we have here
                    // 0=IO-Pins ; 1= SpecialModes
                    case IOWarrior.PID_IOW24:
                    case IOWarrior.PID_IOW24PV_1:
                    case IOWarrior.PID_IOW24PV_2:
                        if(hidCaps.InputReportByteLength==3)
                            interfaceNumber=0;
                        else
                            interfaceNumber=1;
                        break;
                    case IOWarrior.PID_IOW40:
                        if(hidCaps.InputReportByteLength==5)
                            interfaceNumber=0;
                        else
                            interfaceNumber=1;
                        break;
                    case IOWarrior.PID_IOW56:
                        if(hidCaps.InputReportByteLength==8)
                            interfaceNumber=0;
                        else
                            interfaceNumber=1;
                        break;
                }
                //Get the serialnumber of the warrior
                IntPtr snPtr = Marshal.AllocHGlobal(32);
                bresult=HidDevice.HidD_GetSerialNumberString(fd, snPtr, 32);
                if(!bresult) {
                    //reading the serial failed, no serial no support
                    Marshal.FreeHGlobal(snPtr);
                    HidDevice.CloseHandle(fd);
                    continue;
                }
                String sn=Marshal.PtrToStringAuto(snPtr);
                serialNumber=uint.Parse(sn, System.Globalization.NumberStyles.HexNumber);
                Marshal.FreeHGlobal(snPtr);
                //test wether we accept this device
                if(!dfilter.accept(product,serialNumber)) {
                    //we don't want this one
                    HidDevice.CloseHandle(fd);
                    continue;
                }
                //we have all the data we need for one of the interfaces 
                //of an IOWarrior
                addInterface(fd, (String)hidPathList[i], product,
                        serialNumber, revision, interfaceNumber);
            }
        }

        /// <summary>
        /// Sets up the IOWarrior devices. For each IOWarrior we have 2 
        /// interfaces. If the specific product/serial combination is not
        /// found in our list of devices, this is a new IOWarrior to be created.
        /// Otherwise this is simply the second interface of the IOWarrior
        /// anmd we are ready to report this IOWarrior to the listener for
        /// device changes.
        /// </summary>
        /// <param name="fd">The internal filedescriptor of the interface</param>
        /// <param name="fn">The filename of the interface</param>
        /// <param name="product">The produt found on the interface</param>
        /// <param name="serial">The serialnumber fond on the interface</param>
        /// <param name="revision">The revisionnumber found on the interface</param>
        /// <param name="interfaceNumber">if 0 its the IO-Pins, if 1 its the
        /// SpecialMode interface.
        /// </param>
        private static void addInterface(int fd, String fn, int product,
             uint serial, int revision, int interfaceNumber) {
            bool found=false;
            IOWarrior.DeviceState status;
            lock(deviceList) {
                IOWarrior iow=null;
                for(int i=0;i<deviceList.Count;i++) {
                    iow=(IOWarrior)deviceList[i];
                    if(iow.matchDevice(product, serial) && iow.isConnected()) {
                        //we already have a one of the interfaces
                        found=true;
                        status=iow.addInterface(fn, fd, product, serial, revision, interfaceNumber);
                        if(status==IOWarrior.DeviceState.RUNNING) {
                            //all threads are up on this IOWarrior
                            //try to get the initial iostatus.
                            byte[] buffer=new byte[8];
                            if(HidDevice.HidD_GetInputReport(iow.getFileDescriptorIO(),
                                buffer, iow.getReportSizeIO())) {
                                iow.reportUpdate(new IowReport(buffer));
                            }
                            //send report for the running device
                            fireDeviceChange(MessageCatcher.DEVICE_ADD,iow);
                        }
                        else if(status==IOWarrior.DeviceState.STARTUP_ERROR) {
                            //this device cannot be used, throw it away (silently)
                            deviceList.RemoveAt(i);
                        }
                    }
                }
                if(!found) {
                    //this is the first interface of a new device 
                    iow=new IOWarrior();
                    iow.addInterface(fn, fd, product, serial, revision, interfaceNumber);
                    deviceList.Add(iow);
                }
            }
        }


        /// <summary>
        /// Register for WM_DEVICECHANGE messages
        /// </summary>
        private static void addDeviceChangeCatcher() {
            int mcHandle=mc.Handle.ToInt32();
            IntPtr bcBuffer;
            int size;
            HidDevice.DEV_BROADCAST_DEVICEINTERFACE bcInterface;
            System.Guid hidGuid=new Guid();

            bcInterface=new HidDevice.DEV_BROADCAST_DEVICEINTERFACE();
            try {
                size = Marshal.SizeOf(bcInterface);
                bcInterface.dbcc_size=size;
                // Request to receive notifications about a class of devices.
                bcInterface.dbcc_devicetype = HidDevice.DBT_DEVTYP_DEVICEINTERFACE;
                bcInterface.dbcc_reserved = 0;
                // Specify the interface class to receive notifications about.
                HidDevice.HidD_GetHidGuid(ref hidGuid);
                bcInterface.dbcc_classguid=hidGuid;

                // Allocate memory for the buffer that holds the DEV_BROADCAST_DEVICEINTERFACE structure.
                bcBuffer = Marshal.AllocHGlobal(size);

                // Copy the DEV_BROADCAST_DEVICEINTERFACE structure to the buffer.
                // Set fDeleteOld True to prevent memory leaks.
                Marshal.StructureToPtr(bcInterface, bcBuffer, true);
                deviceNotifyHandle=HidDevice.RegisterDeviceNotification(mc.Handle,
                    bcBuffer, HidDevice.DEVICE_NOTIFY_WINDOW_HANDLE);
                // Marshal data from the unmanaged block bcBuffer to
                // the managed object DevBroadcastDeviceInterface
                //Marshal.PtrToStructure(bcBuffer, bcInterface);

                // Free the memory allocated previously by AllocHGlobal.
                Marshal.FreeHGlobal(bcBuffer);
                if(deviceNotifyHandle.ToInt32() == IntPtr.Zero.ToInt32()) {
                    throw new System.ComponentModel.Win32Exception();
                }

            }
            catch(Exception ex) {
                //this here is a really stupid construct
                //for supressing a warning from the compiler about
                //an unused variable
                if(ex!=null)
                    ex=null;
                //End of stupid construct
                throw new IowManager.IowManagerException(new IowError(IowError.Type.LIB_ERROR,
                    null,"Unable to register for device changes",0));
            }
        }

        /// <summary>
        /// Unregister ourselfes from thw WM_DEVICECHANGE messages
        /// </summary>
        private static void removeDeviceChangeCatcher() {
            try {
                HidDevice.UnregisterDeviceNotification(deviceNotifyHandle);

            }
            catch(Exception ex) {
                Console.WriteLine("something wrong in messagecatcher DEregister");
                Console.WriteLine(ex.ToString());
            }
        }

        /// <summary>
        /// internal class for device notifications. The run() method
        /// of this class will be started in a thread of its own.
        /// </summary>
        private class DeviceNotifyThread {
            private IowDeviceChangeListener listener;
            private int action;
            private IOWarrior iow;

            public DeviceNotifyThread(IowDeviceChangeListener listener, int action, IOWarrior iow) {
                this.listener=listener;
                this.action=action;
                this.iow=iow;
            }

            public void run() {
                if(listener!=null) {
                    if(action==MessageCatcher.DEVICE_ADD)
                        listener.deviceAdd(iow);
                    else if(action==MessageCatcher.DEVICE_REMOVED)
                        listener.deviceRemoved(iow);
                }
            }
        }

        /// <summary>
        /// An internal exception thrown during device enumeration.
        /// </summary>
        internal class IowManagerException : Exception {
            private IowError error;

            internal IowManagerException(IowError error) {
                this.error=error;
            }

            internal IowError getError() {
                return error;
            }
        }

    }
}
