using System;
using System.Threading;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace org.wayoda.csiow {
    /// <summary>
    /// A class for a thread that reads from one of the interfaces
    /// of an iowarrior
    /// </summary>
    internal sealed class IowReadThread {
        /// <summary>
        /// Constant for the IO-PIN interface of an IOWarrior
        /// </summary>
        internal const int INTERFACE_IO=0;
        /// <summary>
        /// Constant for the SpecialMode interface of an IOWarrior
        /// </summary>
        internal const int INTERFACE_SM=1;

        //the device for which we do our work
        private IOWarrior iow=null;
        //the interface from which we read
        private int iowInterface;
        //An event that is triggered when the thread needs to shutdown
        private IntPtr shutdownEvent;
        //An event that is triggred when we read something or had a timeout
        private IntPtr readEvent;
        //an array of events to be monitored
        private IntPtr[] events=new IntPtr[2];
        //an overlapped structure for reading
        private NativeOverlapped ovlp;

        /// <summary>
        /// Creates a new thread that reads from one of the interfaces
        /// of an IOWarrior
        /// </summary>
        /// <param name="iow">
        /// The device for which this thread is started
        /// </param>
        /// <param name="iowInterface">
        /// The interface which we monitor in this thread. This must be
        /// one of the constants IowReadThread.INTERFACE_IO or
        /// IowReadThread.INTERFACE_SM
        /// </param>
        internal IowReadThread(IOWarrior iow, int iowInterface) {
            this.iow=iow;
            if(iowInterface!=INTERFACE_IO && iowInterface!=INTERFACE_SM)
                throw new ArgumentException("Invalid interface ["+iowInterface+"]");
            this.iowInterface=iowInterface;
            HidDevice.SECURITY_ATTRIBUTES sec=new HidDevice.SECURITY_ATTRIBUTES();
            sec.lpSecurityDescriptor = 0;
            sec.bInheritHandle = System.Convert.ToInt32(true);
            sec.nLength = Marshal.SizeOf(sec);
            shutdownEvent=new IntPtr(HidDevice.CreateEvent(ref sec, 0, 0, ""));
            if(shutdownEvent.ToInt32()==IntPtr.Zero.ToInt32()) {
                Console.WriteLine("create shutdownEvent failed");
                return;
            }
            readEvent=new IntPtr(HidDevice.CreateEvent(ref sec, 0, 1, ""));
            if(readEvent.ToInt32()==IntPtr.Zero.ToInt32()) {
                Console.WriteLine("create readEvent failed");
                return;
            }
            events[0]=shutdownEvent;
            events[1]=readEvent;
            ovlp=new NativeOverlapped();
            ovlp.EventHandle=readEvent;
            ovlp.OffsetHigh=0;
            ovlp.OffsetLow=0;
        }

        /// <summary>
        /// Notifies the thread to stop reading. The thread-routine will do
        /// some cleanup and then terminate.
        /// </summary>
        internal void shutdown() {
            if(shutdownEvent!=IntPtr.Zero)
                HidDevice.SetEvent(shutdownEvent);
        }

        /// <summary>
        /// The method that is run by the thread.
        /// </summary>
        internal void run() {
            uint ioResult;
            int readResult;
            int fd;
            int size=0;
            int bytesRead=0;
            //the buffer into which we read our data
            byte[] buffer;

            try {
                if(iowInterface==INTERFACE_IO) {
                    fd=iow.getFileDescriptorIO();
                    size=iow.getReportSizeIO();
                }
                else {
                    fd=iow.getFileDescriptorSM();
                    size=iow.getReportSizeSM();
                }
                buffer=new byte[size];
                while(true) {
                    bytesRead=0;
                    readResult=HidDevice.ReadFile(fd, buffer, size, ref bytesRead, ref ovlp);
                    if(readResult==0 && Marshal.GetLastWin32Error()==HidDevice.ERROR_HANDLE_EOF) {
                        //Device was unplugged, we die silently. Apps will be notified
                        //by the devicechange event
                        break;
                    }
                    ioResult=HidDevice.WaitForMultipleObjects(2, events, false, 5000);
                    if(ioResult==HidDevice.WAIT_OBJECT_0) {
                        //shutdown from the IowManager (device was unplugged or close called)
                        HidDevice.CancelIo(fd);
                        break;
                    }
                    else if(ioResult==HidDevice.WAIT_OBJECT_0+1) {
                        if((!HidDevice.GetOverlappedResult(fd, ref ovlp, ref bytesRead, false))
                        && Marshal.GetLastWin32Error()==HidDevice.ERROR_HANDLE_EOF) {
                            //device was unplugged
                            break;
                        }
                        else if(bytesRead==size) {
                            //read was ok
                            IowReport rep=new IowReport(buffer);
                            iow.reportUpdate(rep);
                        }
                    }
                    else if(ioResult==HidDevice.WAIT_TIMEOUT) {
                        //we timed out and go again
                        HidDevice.CancelIo(fd);
                    }
                    else {
                        //some strange error happend here
                        Console.WriteLine("some error="+ioResult);
                        HidDevice.CancelIo(fd);
                        break;
                    }
                }
            }
            finally {
                //do this even if we are aborted
                HidDevice.CloseHandle(readEvent.ToInt32());
                HidDevice.CloseHandle(shutdownEvent.ToInt32());
                readEvent=IntPtr.Zero;
                shutdownEvent=IntPtr.Zero;
            }
        }
    }
}
