using System;
using System.Collections.Generic;
using System.Text;

namespace org.wayoda.csiow {
    /// <summary>
    /// A class for data read from/written to IOWarriors.
    /// An IowReport is an abstraction from the different
    /// reportsizes of the IOWarriors and the concept of the
    /// IO-Pin vs. SpecialMode functions. Users of this library
    /// will only send and receive IowReport.
    /// </summary>
    public sealed class IowReport {
        /// <summary>
        /// The maximum size of data supported by any IOWarrior.
        /// This is also the maximum number to be used for the 
        /// index-operator if the IowReport.
        /// </summary>
        public const int PAYLOAD_SIZE=63;
        /// <summary>
        /// the actual data for the report
        /// </summary>
        private byte[] data;

        /// <summary>
        /// Creates a new IowReport with a report-id of 0x00
        /// and all data elements set to 0x00
        /// </summary>
        public IowReport() {
            data=new byte[PAYLOAD_SIZE+1];
        }

        /// <summary>
        /// Creates a new IowReport from an array of byte
        /// </summary>
        /// <param name="data">
        /// The data for the report. The report-id is expected in
        /// data[0] and the payload in the rest of the array. Internally
        /// we will simply copy the elements into the report.
        /// </param>
        public IowReport(byte[] data) {
            this.data=new byte[PAYLOAD_SIZE+1];
            if(data!=null) {
                for(int i=0;i<data.Length && i<PAYLOAD_SIZE;i++)
                    this.data[i]=data[i];
            }
        }

        /// <summary>
        /// Creates a new IowReport with a specific report-id and
        /// all the data-elements set to 0x00.
        /// </summary>
        /// <param name="reportId">The report-id to be used with
        /// this report</param>
        public IowReport(byte reportId) {
            data=new byte[PAYLOAD_SIZE+1];
            data[0]=reportId;
        }

        /// <summary>
        /// Creates a new IowReport with a specific report-id and
        /// the data taken from an array of byte.
        /// </summary>
        /// <param name="reportId">
        /// The report-id to be used with this report.
        /// </param>
        /// <param name="data">
        /// The data for this report.
        /// </param>
        public IowReport(byte reportId, byte[] data) {
            this.data=new byte[PAYLOAD_SIZE+1];
            this.data[0]=reportId;
            if(data!=null) {
                for(int i=0;i<data.Length && i<PAYLOAD_SIZE-1;i++)
                    this.data[i+1]=data[i];
            }
        }

        /// <summary>
        /// Creates a deep copy of an IowReport.
        /// </summary>
        /// <param name="rep">
        /// The report to be copied.
        /// </param>
        public IowReport(IowReport rep) {
            if(rep!=null) {
                this.data=rep.getInternalArray();
            }
        }

        /// <summary>
        /// Tests wether two reports are equal.
        /// </summary>
        /// <param name="o">The object to be tested for equality</param>
        /// <returns>True if the obecjt passed in is an 
        /// IowReport with exactly the same report-id and dataelements
        /// as this one.
        /// </returns>
        public override bool Equals(Object o) {
            IowReport r=(o as IowReport);
            if(r==null)
                return false;
            if(r==this)
                return true;
            byte[] data=r.getInternalArray();
            for(int i=0;i<this.data.Length;i++) {
                if(this.data[i]!=data[i])
                    return false;
            }
            return false;
        }

        /// <summary>
        /// Gets the hashcode for this report
        /// </summary>
        /// <returns>Returns the hashcode for this report</returns>
        public override int GetHashCode() {
            return base.GetHashCode();
        }

        /// <summary>
        /// Gets a deep copy of the report.
        /// </summary>
        /// <returns>
        /// An array of byte that is a deep copy of this reports data.
        /// </returns>
        internal byte[] getInternalArray() {
            byte[] retval=new byte[PAYLOAD_SIZE+1];
            data.CopyTo(retval, 0);
            return retval;
        }

        /// <summary>
        /// Gets the report-id of the report.
        /// </summary>
        /// <returns>
        /// the report-id for this report
        /// </returns>
        public byte getReportId() {
            return data[0];
        }

        /// <summary>
        /// Gets the IowSpecialMode group for this report.
        /// </summary>
        /// <returns>
        /// One of the typed constants for IowSpecialModes, or null
        /// if the report-id does not address a valid special-mode.
        /// </returns>
        public IowSpecialMode getSpecialMode() {
            IowSpecialMode m=IowSpecialMode.getByReportId(data[0]);
            return m;
        }

        /// <summary>
        /// Gets the data in this report. The report-id is 
        /// not included into the array returned.
        /// </summary>
        /// <returns>
        /// Gets an array of byte that contains only the data
        /// in this report. The value found at array[0] is the
        /// first byte of data, NOT the report-id.
        /// </returns>
        public byte[] getReportData() {
            byte[] retval=new byte[PAYLOAD_SIZE]; 
            for(int i=0,j=1;i<PAYLOAD_SIZE;i++,j++)
                retval[i]=this.data[j];
            return data;
        }

        /// <summary>
        /// Implements the index operator for IowReports. Supports
        /// get and set.
        /// </summary>
        /// <param name="index">The index into the report</param>
        /// <returns>the value at index in a get operation</returns>
        public byte this[int index] {
            get {
                if(index<0 || (index > this.data.Length-1))
                    throw new IndexOutOfRangeException("Invalid index for IowReport ("+index+")");
                return this.data[index];
            }
            set {
                if(index<0 || (index > this.data.Length-1))
                    throw new IndexOutOfRangeException("Invalid index for IowReport ("+index+")");
                this.data[index]=value;
            }
        }

        /// <summary>
        /// Copies a report into this one. 
        /// </summary>
        /// <param name="rep">The report to copy</param>
        public void set(IowReport rep) {
            if(rep!=null)
                this.data=rep.getInternalArray();
        }

        /// <summary>
        /// Set the report-id and data for this report.
        /// </summary>
        /// <param name="reportId">The report-id to be used</param>
        /// <param name="data">The data to be used. It is not
        /// required that the length data must match the payload-size
        /// of a report. Well will simply copy whats available.
        /// </param>
        public void set(byte reportId, byte[] data) {
            this.data=new byte[PAYLOAD_SIZE+1];
            this.data[0]=reportId;
            if(data!=null) {
                for(int i=0;i<PAYLOAD_SIZE-1;i++) {
                    if(i<data.Length)
                        this.data[i+1]=data[i];
                    this.data[i+1]=0x00;
                }
            }
        }

        /// <summary>
        /// Gets a string representation of this report.
        /// The string is formatted into multiple lines.
        /// Use for debugging only.
        /// </summary>
        /// <returns>
        /// A String with the contents of this report
        /// </returns>
        public override string ToString() {
            StringBuilder sb=new StringBuilder();
            sb.Append("ReportId = ");
            sb.AppendFormat("{0:X2}", data[0]);
            for(int i=0;i<PAYLOAD_SIZE;i++) {
                if((i%16)==0)
                    sb.Append("\r\n");
                sb.AppendFormat("{0:X2} ", data[i+1]);
            }
            sb.Append("\r\n");
            return sb.ToString();
        }
    }
}
