using System;
using System.Collections.Generic;
using System.Text;

namespace org.wayoda.csiow {
    /// <summary>
    /// A collection of typed constants for all SpecialModes
    /// of an IOWarrior. The IO-Pin function is treated here
    /// as just another SpecialMode.
    /// </summary>
    public sealed class IowSpecialMode {
        /// <summary>
        /// A typed constant for the IO-Pin mode.
        /// </summary>
        public static readonly IowSpecialMode IO=new IowSpecialMode(0x00, "IO");
        /// <summary>
        /// A typed constant for the IIC mode.
        /// </summary>
        public static readonly IowSpecialMode IIC=new IowSpecialMode(0x01, "IIC");
        /// <summary>
        /// A typed constant for the LCD mode.
        /// </summary>
        public static readonly IowSpecialMode LCD=new IowSpecialMode(0x04, "LCD");
        /// <summary>
        /// A typed constant for the SPI mode.
        /// </summary>
        public static readonly IowSpecialMode SPI=new IowSpecialMode(0x08, "SPI");
        /// <summary>
        /// A typed constant for the RC5 (remote control) mode.
        /// </summary>
        public static readonly IowSpecialMode RC5=new IowSpecialMode(0x0C, "RC5");
        /// <summary>
        /// A typed constant for the Led Matrix mode.
        /// </summary>
        public static readonly IowSpecialMode LED_MATRIX=new IowSpecialMode(0x14, "Led-Matrix");
        /// <summary>
        /// A typed constant for the Switch Matrix mode.
        /// </summary>
        public static readonly IowSpecialMode SWITCH_MATRIX=new IowSpecialMode(0x18, "Switch-Matrix");
        /// <summary>
        /// A typed constant for the PowerStatus function on an IOWarrior24PV.
        /// </summary>
        public static readonly IowSpecialMode POWER_STATUS=new IowSpecialMode(0x20, "Power-Status");        
        /// <summary>
        /// A typed constant for the GetCurrentPinsStatus mode.
        /// </summary>
        public static readonly IowSpecialMode PIN_STATUS=new IowSpecialMode(0xFF, "IO-Status");
        
        // The report-ids that belong to a specific special-mode
        private byte[] reportIds;
        //the name of the mode
        private String name;

        /// <summary>
        /// Constructor is private because we want to have only
        /// static inatnces of this class
        /// </summary>
        /// <param name="modeId">
        /// The report-id that switches this mode onor off
        /// </param>
        /// <param name="name">
        /// A name for the mode.
        /// </param>
        private IowSpecialMode(byte modeId, String name) {
            if(modeId==0x00) {
                byte[] ri={ 0x00 };
                reportIds=ri;
            }
            else if(modeId==0x01) {
                byte[] ri={ 0x01, 0x02, 0x03 };
                reportIds=ri;
            }
            else if(modeId==0x04) {
                byte[] ri={ 0x04, 0x05, 0x06 };
                reportIds=ri;
            }
            else if(modeId==0x08) {
                byte[] ri={ 0x08, 0x09 };
                reportIds=ri;
            }
            else if(modeId==0x0C) {
                byte[] ri={ 0x0C };
                reportIds=ri;
            }
            else if(modeId==0x14) {
                byte[] ri={ 0x14, 0x15 };
                reportIds=ri;
            }
            else if(modeId==0x18) {
                byte[] ri={ 0x18, 0x19, 0x1A };
                reportIds=ri;
            }
            else if(modeId==0x20) {
                byte[] ri={ 0x20, 0x21 };
                reportIds=ri;
            }
            else if(modeId==0xFF) {
                byte[] ri={ 0xFF };
                reportIds=ri;
            }
            this.name=name;
        }
        
        public override int GetHashCode() {
            return (int)reportIds[0];
        }


        public override bool Equals(object o) {
            IowSpecialMode m=(o as IowSpecialMode);
            return this==m;
        }

        /// <summary>
        /// Returns an IowSpecialMode for a reportd-id. 
        /// Example: the report-ids (0x18, 0x19, 0x1A) all belong
        /// to the SwitchMatrix-mode so for each of them we return
        /// IowSpecialMode.SWITCH_MATRIX.
        /// If you use an invalid report-id which does not belong
        /// to a specialmode we return null.
        /// </summary>
        /// <param name="id">
        /// The report-id for which the specialmode is requested
        /// </param>
        /// <returns>
        /// The IowSpecialMode to which the report-id belongs.
        /// </returns>
        public static IowSpecialMode getByReportId(byte id) {
            switch(id) {
                case 0x00:
                    return IO;
                case 0x01:
                case 0x02:
                case 0x03:
                    return IIC;
                case 0x04:
                case 0x05:
                case 0x06:
                    return LCD;
                case 0x08:
                case 0x09:
                    return SPI;
                case 0x0C:
                    return RC5;
                case 0x14:
                case 0x15:
                    return LED_MATRIX;
                case 0x18:
                case 0x19:
                case 0x1A:
                    return SWITCH_MATRIX;
                case 0x20 :
                case 0x21 :
                    return POWER_STATUS;
                case 0xFF:
                    return PIN_STATUS;
            }
            return null;
        }

        /// <summary>
        /// Tests wether this IowSpecialMode understands a specific
        /// report-id
        /// </summary>
        /// <param name="id">The report-id to be tested</param>
        /// <returns>True if the report-id is understood by this specialmode</returns>
        public bool isSupportedId(byte id) {
            for(int i=0;i< reportIds.Length;i++) {
                if(reportIds[i]==id)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Gets the report-id to enable/disable this specialmode.
        /// </summary>
        /// <returns>The report-id to be used for enabling or disabling
        /// this specialmode
        /// </returns>
        public byte getEnableReportId() {
            return reportIds[0];
        }

        /// <summary>
        /// Gets the name of this specialmode.
        /// </summary>
        /// <returns>A String with the name of this specialmode</returns>
        public override String ToString() {
            return name;
        }


    }
}
