using System;
using System.Threading;
using System.IO;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Text;

namespace org.wayoda.csiow {
    /// <summary>
    /// This class writes reports to an iowarrior using overlapped io.
    /// Writes can be canceled if they take to long.
    /// </summary>
    internal sealed class IowWriter {
        //the device for which we do our work
        private IOWarrior iow=null;
        //An event that is triggered when the thread needs to shutdown
        private IntPtr shutdownEvent;
        //An event that is triggered when the write timed out
        private IntPtr writeEvent;
        //an array of events to be monitored by writefile
        private IntPtr[] events=new IntPtr[2];
        //an overlapped structure for writing
        private NativeOverlapped ovlp;
        //true if setup for the writer was ok
        private bool setup=false;

        internal IowWriter(IOWarrior iow) {
            this.iow=iow;
            HidDevice.SECURITY_ATTRIBUTES sec=new HidDevice.SECURITY_ATTRIBUTES();
            sec.lpSecurityDescriptor = 0;
            sec.bInheritHandle = System.Convert.ToInt32(true);
            sec.nLength = Marshal.SizeOf(sec);
            shutdownEvent=new IntPtr(HidDevice.CreateEvent(ref sec, 0, 0, ""));
            if(shutdownEvent.ToInt32()==IntPtr.Zero.ToInt32()) {
                return;
            }
            writeEvent=new IntPtr(HidDevice.CreateEvent(ref sec, 0, 1, ""));
            if(writeEvent.ToInt32()==IntPtr.Zero.ToInt32()) {
                return;
            }
            events[0]=shutdownEvent;
            events[1]=writeEvent;
            ovlp=new NativeOverlapped();
            ovlp.EventHandle=writeEvent;
            ovlp.OffsetHigh=0;
            ovlp.OffsetLow=0;
            setup=true;
        }

        internal void shutdown() {
            HidDevice.CloseHandle(writeEvent.ToInt32());
            HidDevice.CloseHandle(shutdownEvent.ToInt32());
            setup=false;
        }

        internal bool isSetup() {
            return setup;
        }

        internal IowError write(IowReport rep) {
            int fd;
            int size;
            int bytesWritten;
            int writeResult;
            uint ioResult;
            int winError;

            if(!setup)
                return new IowError(IowError.Type.WRITE_FAILED, iow,
                    "Internal write error (probably unplugged)", 0);
            byte[] buffer=rep.getInternalArray();
            if(rep.getReportId()==0x00) {
                fd=iow.getFileDescriptorIO();
                size=iow.getReportSizeIO();
            }
            else {
                fd=iow.getFileDescriptorSM();
                size=iow.getReportSizeSM();
            }
            bytesWritten=0;
            writeResult=HidDevice.WriteFile(fd, buffer, size, ref bytesWritten, ref ovlp);
            if(writeResult==0 && Marshal.GetLastWin32Error()==HidDevice.ERROR_HANDLE_EOF) {
                //Device was unplugged, we die silently. Apps will be notified
                //by the devicechange event
                return new IowError(IowError.Type.UNPLUGGED, iow,
                    "Device is unplugged", 0);
            }
            //we wait for 5 secs before we timeout
            ioResult=HidDevice.WaitForMultipleObjects(2, events, false, 5000);
            if(ioResult==HidDevice.WAIT_OBJECT_0) {
                //shutdown from the IowManager (device was unplugged or close called)
                HidDevice.CancelIo(fd);
                return new IowError(IowError.Type.UNPLUGGED, iow,
                    "Device is unplugged", 0);
            }
            else if(ioResult==HidDevice.WAIT_OBJECT_0+1) {
                if((!HidDevice.GetOverlappedResult(fd, ref ovlp, ref bytesWritten, false))
                        && Marshal.GetLastWin32Error()==HidDevice.ERROR_HANDLE_EOF) {
                    //device was unplugged
                    return new IowError(IowError.Type.UNPLUGGED, iow,
                        "Device is unplugged", 0);
                }
                else if(bytesWritten==size) {
                    return IowError.OK;
                }
            }
            else if(ioResult==HidDevice.WAIT_TIMEOUT) {
                //we timed out
                HidDevice.CancelIo(fd);
                return new IowError(IowError.Type.WRITE_TIMEOUT, iow,
                    "Timeout in write >= 5 seconds", 0);
            }
            else if(ioResult==HidDevice.WAIT_FAILED) {
                winError=Marshal.GetLastWin32Error();
                HidDevice.CancelIo(fd);
                return new IowError(IowError.Type.WRITE_FAILED, iow,
                    "Internal write error (WAIT_FAILED)", winError);
            }
            //some strange error happend here
            winError=Marshal.GetLastWin32Error();
            HidDevice.CancelIo(fd);
            return new IowError(IowError.Type.WRITE_FAILED, iow,
                "Internal write error", winError);
        }
    }
}
